from functools import reduce

import numpy as np
from astropy.wcs.wcsapi.wrappers.base import BaseWCSWrapper

__all__ = ['CompoundLowLevelWCS']


def tuplesum(lists):
    return reduce(tuple.__add__, map(tuple, lists))


class Mapping:
    """
    Allows inputs to be reordered, duplicated or dropped.

    This is a very stripped down version of `astropy.modeling.models.Mapping`
    to be able to handle input of arbitrary type.

    Parameters
    ----------
    mapping : tuple
        A tuple of integers representing indices of the inputs to this model
        to return and in what order to return them.  See
        :ref:`compound-model-mappings` for more details.

    """
    def __init__(self, mapping):
        self.mapping = mapping
        self.n_inputs = max(mapping) + 1
        self.n_outputs = len(mapping)

    def __call__(self, *values):
        return tuple(values[idx] for idx in self.mapping)

    @property
    def inverse(self):
        mapping = tuple(self.mapping.index(idx)
                        for idx in range(self.n_inputs))
        return type(self)(mapping)

    def __repr__(self):
        return f'<Mapping({self.mapping})>'


class CompoundLowLevelWCS(BaseWCSWrapper):
    """
    A wrapper that takes multiple low level WCS objects and makes a compound
    WCS that combines them.

    Parameters
    ----------
    *wcs : `~astropy.wcs.wcsapi.BaseLowLevelWCS`
        The WCSes to combine
    mapping : `tuple`
        The pixel dimension mapping between the input pixel dimensions and the
        input pixel dimensions to the underlying WCSes. This should have length
        equal to the total number of pixel dimensions in all input WCSes and
        have a maximum of the number of input pixel dimensions to the resulting
        compound WCS -1 (counts from 0). For example ``(0, 1, 2, 1)`` would end
        up with the second and fourth pixel dimensions in the input WCSes being
        shared, so the compound WCS would have 3 pixel dimensions ``(2 + 1)``.
        See :ref:`compound-model-mappings` for more examples of this input
        format.
    pixel_atol : `float`
        A tolerance used to check that the resulting pixel coordinates from
        ``world_to_pixel`` are the same from all WCSes.
    """
    def __init__(self, *wcs, mapping=None, pixel_atol=1e-8):
        self._wcs = wcs

        if not mapping:
            mapping = tuple(range(self._all_pixel_n_dim))

        if not len(mapping) == self._all_pixel_n_dim:
            raise ValueError(
                "The length of the mapping must equal the total number of pixel dimensions in all input WCSes.")

        self.mapping = Mapping(mapping)
        self.atol = pixel_atol

        # Validate the pixel bounds and shape are consistent
        self.pixel_bounds
        self.pixel_shape

    @property
    def _all_pixel_n_dim(self):
        return sum([w.pixel_n_dim for w in self._wcs])

    @property
    def pixel_n_dim(self):
        return self.mapping.n_inputs

    @property
    def world_n_dim(self):
        return sum([w.world_n_dim for w in self._wcs])

    @property
    def world_axis_physical_types(self):
        return tuplesum([w.world_axis_physical_types for w in self._wcs])

    @property
    def world_axis_units(self):
        return tuplesum([w.world_axis_units for w in self._wcs])

    def pixel_to_world_values(self, *pixel_arrays):
        pixel_arrays = self.mapping(*pixel_arrays)
        world_arrays = []
        for w in self._wcs:
            pixel_arrays_sub = pixel_arrays[:w.pixel_n_dim]
            pixel_arrays = pixel_arrays[w.pixel_n_dim:]
            world_arrays_sub = w.pixel_to_world_values(*pixel_arrays_sub)
            if w.world_n_dim > 1:
                world_arrays.extend(world_arrays_sub)
            else:
                world_arrays.append(world_arrays_sub)
        return tuple(world_arrays)

    def world_to_pixel_values(self, *world_arrays):
        pixel_arrays = []
        for w in self._wcs:
            world_arrays_sub = world_arrays[:w.world_n_dim]
            world_arrays = world_arrays[w.world_n_dim:]
            pixel_arrays_sub = w.world_to_pixel_values(*world_arrays_sub)
            if w.pixel_n_dim > 1:
                pixel_arrays.extend(pixel_arrays_sub)
            else:
                pixel_arrays.append(pixel_arrays_sub)

        mapped_axes = set(self.mapping.mapping)
        for mapped_axis in mapped_axes:
            idx, = np.atleast_1d(self.mapping.mapping == mapped_axis).nonzero()
            if len(idx) > 1:
                idx_0 = idx[0]
                for idx_n in idx[1:]:
                    if not np.allclose(pixel_arrays[idx_0], pixel_arrays[idx_n],
                                       atol=self.atol, equal_nan=True):
                        raise ValueError(
                            "The world inputs for shared pixel axes did not result in a pixel "
                            f"coordinate to within {self.atol} relative accuracy."
                        )
        return self.mapping.inverse(*pixel_arrays)

    @property
    def world_axis_object_components(self):
        all_components = []
        for iw, w in enumerate(self._wcs):
            for component in w.world_axis_object_components:
                all_components.append((f'{component[0]}_{iw}',) + component[1:])
        return all_components

    @property
    def world_axis_object_classes(self):
        # TODO: deal with name conflicts
        all_classes = {}
        for iw, w in enumerate(self._wcs):
            for key, value in w.world_axis_object_classes.items():
                all_classes[f'{key}_{iw}'] = value
        return all_classes

    @property
    def pixel_shape(self):
        if not any(w.array_shape is None for w in self._wcs):
            pixel_shape = tuplesum(w.pixel_shape for w in self._wcs)
            out_shape = self.mapping.inverse(*pixel_shape)
            for i, ix in enumerate(self.mapping.mapping):
                if out_shape[ix] != pixel_shape[i]:
                    raise ValueError(
                        "The pixel shapes of the supplied WCSes do not match for the dimensions shared by the supplied mapping.")
            return out_shape

    @property
    def pixel_bounds(self):
        if not any(w.pixel_bounds is None for w in self._wcs):
            pixel_bounds = tuplesum(w.pixel_bounds for w in self._wcs)
            out_bounds = self.mapping.inverse(*pixel_bounds)
            for i, ix in enumerate(self.mapping.mapping):
                if out_bounds[ix] != pixel_bounds[i]:
                    raise ValueError(
                        "The pixel bounds of the supplied WCSes do not match for the dimensions shared by the supplied mapping.")
            return out_bounds

    @property
    def pixel_axis_names(self):
        pixel_names = tuplesum(w.pixel_axis_names for w in self._wcs)
        out_names = self.mapping.inverse(*pixel_names)

        for i, ix in enumerate(self.mapping.mapping):
            if out_names[ix] != pixel_names[i]:
                out_names[ix] = ' / '.join([out_names[ix], pixel_names[i]])

        return out_names

    @property
    def world_axis_names(self):
        return tuplesum(w.world_axis_names for w in self._wcs)

    @property
    def axis_correlation_matrix(self):
        full_matrix = np.zeros((self.world_n_dim, self._all_pixel_n_dim), dtype=bool)
        iw = ip = 0
        for w in self._wcs:
            full_matrix[iw:iw + w.world_n_dim, ip:ip + w.pixel_n_dim] = w.axis_correlation_matrix
            iw += w.world_n_dim
            ip += w.pixel_n_dim

        matrix = np.zeros((self.world_n_dim, self.pixel_n_dim), dtype=bool)
        for i, ix in enumerate(self.mapping.mapping):
            matrix[:, ix] = np.logical_or(matrix[:, ix], full_matrix[:, i])

        return matrix

    @property
    def serialized_classes(self):
        return any([w.serialized_classes for w in self._wcs])



import astropy.units as u
import gwcs.coordinate_frames as cf
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.time import Time

from ndcube.extra_coords.table_coord import (MultipleTableCoordinate, QuantityTableCoordinate,
                                             SkyCoordTableCoordinate, TimeTableCoordinate)


@pytest.fixture
def lut_1d_distance():
    lookup_table = u.Quantity(np.arange(10) * u.km)
    return QuantityTableCoordinate(lookup_table, names='x')


@pytest.fixture
def lut_3d_distance_mesh():
    lookup_table = (u.Quantity(np.arange(10) * u.km),
                    u.Quantity(np.arange(10, 20) * u.km),
                    u.Quantity(np.arange(20, 30) * u.km))

    return QuantityTableCoordinate(*lookup_table, mesh=True, names=['x', 'y', 'z'])


@pytest.fixture
def lut_2d_distance_no_mesh():
    lookup_table = np.arange(9).reshape(3, 3) * u.km, np.arange(9, 18).reshape(3, 3) * u.km
    return QuantityTableCoordinate(*lookup_table, mesh=False)


@pytest.fixture
def lut_1d_skycoord_no_mesh():
    sc = SkyCoord(range(10), range(10), unit=u.deg)
    return SkyCoordTableCoordinate(sc, mesh=False, names=['lon', 'lat'])


@pytest.fixture
def lut_2d_skycoord_no_mesh():
    data = np.arange(9).reshape(3, 3), np.arange(9, 18).reshape(3, 3)
    sc = SkyCoord(*data, unit=u.deg)
    return SkyCoordTableCoordinate(sc, mesh=False)


@pytest.fixture
def lut_2d_skycoord_mesh():
    sc = SkyCoord(range(10), range(10), unit=u.deg)
    return SkyCoordTableCoordinate(sc, mesh=True)


@pytest.fixture
def lut_3d_skycoord_mesh():
    sc = SkyCoord(range(10), range(10), range(10), unit=(u.deg, u.deg, u.AU))
    return SkyCoordTableCoordinate(sc, mesh=True)


@pytest.fixture
def lut_1d_time():
    data = Time(["2011-01-01T00:00:00",
                 "2011-01-01T00:00:10",
                 "2011-01-01T00:00:20",
                 "2011-01-01T00:00:30"], format="isot")
    return TimeTableCoordinate(data, names='time', physical_types='time')


@pytest.fixture
def lut_1d_wave():
    # TODO: Make this into a SpectralCoord object
    return QuantityTableCoordinate(range(10) * u.nm)


def test_exceptions():
    with pytest.raises(TypeError) as ei:
        QuantityTableCoordinate(u.Quantity([1, 2, 3], u.nm), [1, 2, 3])
    assert "All tables must be astropy Quantity objects" in str(ei)

    with pytest.raises(u.UnitsError) as ei:
        QuantityTableCoordinate(u.Quantity([1, 2, 3], u.nm), [1, 2, 3] * u.deg)
    assert "All tables must have equivalent units." in str(ei)

    with pytest.raises(ValueError) as ei:
        QuantityTableCoordinate(u.Quantity([1, 2, 3], u.nm), [1, 2, 3] * u.m, names='x')
    assert "The number of names should match the number of world dimensions" in str(ei)

    with pytest.raises(ValueError) as ei:
        QuantityTableCoordinate(u.Quantity([1, 2, 3], u.nm), [1, 2, 3] * u.m, physical_types='x')
    assert "The number of physical types should match the number of world dimensions" in str(ei)

    # Test two Time
    with pytest.raises(ValueError) as ei:
        TimeTableCoordinate(Time("2011-01-01"), Time("2011-01-01"))
    assert "single Time object" in str(ei)

    with pytest.raises(ValueError) as ei:
        TimeTableCoordinate(Time("2011-01-01"), names=['a', 'b'])
    assert "only have one name." in str(ei)

    with pytest.raises(ValueError) as ei:
        TimeTableCoordinate(Time("2011-01-01"), physical_types=['a', 'b'])
    assert "only have one physical type." in str(ei)

    # Test two SkyCoord
    with pytest.raises(ValueError) as ei:
        SkyCoordTableCoordinate(SkyCoord(10, 10, unit=u.deg), SkyCoord(10, 10, unit=u.deg))
    assert "single SkyCoord object" in str(ei)

    with pytest.raises(ValueError) as ei:
        SkyCoordTableCoordinate(SkyCoord(10, 10, unit=u.deg), names='x')
    assert "names must equal two" in str(ei)

    with pytest.raises(ValueError) as ei:
        SkyCoordTableCoordinate(SkyCoord(10, 10, unit=u.deg), physical_types='x')
    assert "physical types must equal two" in str(ei)

    with pytest.raises(TypeError) as ei:
        MultipleTableCoordinate(10, SkyCoordTableCoordinate(SkyCoord(10, 10, unit=u.deg)))
    assert "All arguments must be BaseTableCoordinate" in str(ei)

    with pytest.raises(TypeError) as ei:
        MultipleTableCoordinate(MultipleTableCoordinate(SkyCoordTableCoordinate(SkyCoord(10, 10, unit=u.deg))))
    assert "All arguments must be BaseTableCoordinate" in str(ei)


def test_1d_distance(lut_1d_distance):
    assert lut_1d_distance.model.n_inputs == 1
    assert lut_1d_distance.model.n_outputs == 1
    assert lut_1d_distance.model.lookup_table.shape == (10,)
    assert u.allclose(u.Quantity(range(10), u.pix), lut_1d_distance.model.points)

    assert u.allclose(lut_1d_distance.wcs.pixel_to_world(0), 0 * u.km)
    assert u.allclose(lut_1d_distance.wcs.pixel_to_world(9), 9 * u.km)
    assert lut_1d_distance.wcs.world_to_pixel(0 * u.km) == 0


def test_3d_distance(lut_3d_distance_mesh):
    ltc = lut_3d_distance_mesh

    assert ltc.model.n_inputs == 3
    assert ltc.model.n_outputs == 3

    assert ltc.wcs.world_n_dim == 3
    assert ltc.wcs.pixel_n_dim == 3

    assert u.allclose(ltc.wcs.pixel_to_world(0*u.pix, 0*u.pix, 0*u.pix),
                      (0, 10, 20)*u.km)
    assert u.allclose(ltc.wcs.world_to_pixel(0*u.km, 10*u.km, 20*u.km), (0, 0, 0))


@pytest.mark.xfail(reason=">1D Tables not supported")
def test_2d_nout_1_no_mesh(lut_2d_distance_no_mesh):
    ltc = lut_2d_distance_no_mesh
    assert ltc.wcs.world_n_dim == 2
    assert ltc.wcs.pixel_n_dim == 2

    assert ltc.model.n_inputs == 2
    assert ltc.model.n_outputs == 2

    assert u.allclose(ltc.wcs.pixel_to_world(0*u.pix, 0*u.pix),
                      (0, 9)*u.km)

    # TODO: this model is not invertable
    assert u.allclose(ltc.wcs.world_to_pixel(0*u.km, 9*u.km), (0, 0))


def test_1d_skycoord_no_mesh(lut_1d_skycoord_no_mesh):
    ltc = lut_1d_skycoord_no_mesh

    assert ltc.model.n_inputs == 1
    assert ltc.model.n_outputs == 2

    pixel_coords = (0,)*u.pix
    sc = ltc.wcs.pixel_to_world(*pixel_coords)
    pix = ltc.wcs.world_to_pixel(sc)
    assert u.allclose(pix, pixel_coords.value)


def test_2d_skycoord_mesh(lut_2d_skycoord_mesh):
    ltc = lut_2d_skycoord_mesh
    assert ltc.model.n_inputs == 2
    assert ltc.model.n_outputs == 2

    pixel_coords = (0, 0)*u.pix
    sc = ltc.wcs.pixel_to_world(*pixel_coords)
    pix = ltc.wcs.world_to_pixel(sc)
    assert u.allclose(pix, pixel_coords.value)


def test_3d_skycoord_mesh(lut_3d_skycoord_mesh):
    ltc = lut_3d_skycoord_mesh

    assert ltc.model.n_inputs == 3
    assert ltc.model.n_outputs == 3

    # Known failure due to gwcs#120

    # pixel_coords = (0, 0, 0)*u.pix
    # sc = ltc.wcs.pixel_to_world(*pixel_coords)
    # pix = ltc.wcs.world_to_pixel(sc)
    # assert u.allclose(pix, pixel_coords.value)

    # assert isinstance(ltc.wcs, gwcs.WCS)
    #
    # sub_ltc = ltc[0:4, 0:5, 0:6]
    # assert sub_ltc.delayed_models[0].lookup_table[0].shape == (4, )
    # assert sub_ltc.delayed_models[0].lookup_table[1].shape == (5, )
    # assert sub_ltc.delayed_models[0].lookup_table[2].shape == (6, )


@pytest.mark.xfail(reason=">1D Tables not supported")
def test_2d_skycoord_no_mesh(lut_2d_skycoord_no_mesh):
    ltc = lut_2d_skycoord_no_mesh

    assert ltc.model.n_inputs == 2
    assert ltc.model.n_outputs == 2

    pixel_coords = (0, 0)*u.pix
    sc = ltc.wcs.pixel_to_world(*pixel_coords)
    pix = ltc.wcs.world_to_pixel(sc)
    assert u.allclose(pix, pixel_coords.value)


def test_1d_time(lut_1d_time):
    assert lut_1d_time.model.n_inputs == 1
    assert lut_1d_time.model.n_outputs == 1
    assert u.allclose(lut_1d_time.model.lookup_table, u.Quantity((0, 10, 20, 30), u.s))

    assert lut_1d_time.wcs.pixel_to_world(0) == Time("2011-01-01T00:00:00")
    assert lut_1d_time.wcs.world_to_pixel(Time("2011-01-01T00:00:00")) == 0


def test_join(lut_1d_time, lut_1d_wave):
    ltc = lut_1d_time & lut_1d_wave

    assert ltc.model.n_inputs == 2
    assert ltc.model.n_outputs == 2

    assert isinstance(ltc.frame, cf.CompositeFrame)
    world = ltc.wcs.pixel_to_world(0, 0)
    assert world[0] == Time("2011-01-01T00:00:00")
    assert u.allclose(world[1], 0 * u.nm)

    assert u.allclose(ltc.wcs.world_to_pixel(*world), (0, 0))


def test_join_3d(lut_2d_skycoord_mesh, lut_1d_wave):
    ltc = lut_2d_skycoord_mesh & lut_1d_wave

    assert ltc.model.n_inputs == 3
    assert ltc.model.n_outputs == 3

    assert isinstance(ltc.frame, cf.CompositeFrame)
    world = ltc.wcs.pixel_to_world(0, 0, 0)
    assert isinstance(world[0], SkyCoord)
    assert u.allclose(world[1], 0 * u.nm)

    pixel_coords = (0, 0, 0)*u.pix
    sc = ltc.wcs.pixel_to_world(*pixel_coords)
    pix = ltc.wcs.world_to_pixel(*sc)
    assert u.allclose(pix, pixel_coords.value)

    assert u.allclose(ltc.wcs.world_to_pixel(*world), (0, 0, 0))


@pytest.mark.xfail(reason=">1D Tables not supported")
def test_2d_quantity():
    shape = (3, 3)
    data = np.arange(np.product(shape)).reshape(shape) * u.m / u.s

    ltc = QuantityTableCoordinate(data)
    assert u.allclose(ltc.wcs.pixel_to_world(0, 0), 0 * u.m / u.s)


def test_repr_str(lut_1d_time, lut_1d_wave):
    assert str(lut_1d_time.table) in str(lut_1d_time)
    assert "TimeTableCoordinate" in repr(lut_1d_time)
    assert str(lut_1d_wave.table)[1:-2] in str(lut_1d_wave)
    assert "QuantityTableCoordinate" in repr(lut_1d_wave)

    join = lut_1d_time & lut_1d_wave
    assert str(lut_1d_time.table) in str(join)
    assert str(lut_1d_wave.table)[1:-2] in str(join)
    assert "TimeTableCoordinate" in repr(join)
    assert "QuantityTableCoordinate" in repr(join)
    assert "MultipleTableCoordinate" in repr(join)


################################################################################
# Slicing Tests
################################################################################


def test_slicing_quantity_table_coordinate():
    qtc = QuantityTableCoordinate(range(10)*u.m, mesh=False, names='x', physical_types='pos:x')

    assert u.allclose(qtc[2:8].table[0], range(2, 8)*u.m)
    assert u.allclose(qtc[2].table[0], 2*u.m)
    assert qtc.names == ['x']
    assert qtc.physical_types == ['pos:x']

    qtc = QuantityTableCoordinate(range(10)*u.m, mesh=True)

    assert u.allclose(qtc[2:8].table[0], range(2, 8)*u.m)
    assert u.allclose(qtc[2].table[0], 2*u.m)

    qtc = QuantityTableCoordinate(range(10)*u.m, range(10)*u.m, mesh=True,
                                  names=['x', 'y'], physical_types=['pos:x', 'pos:y'])
    assert u.allclose(qtc[2:8, 2:8].table[0], range(2, 8)*u.m)
    assert u.allclose(qtc[2:8, 2:8].table[1], range(2, 8)*u.m)

    # we have dropped one dimension
    assert len(qtc[2, 2:8].table) == 1
    assert u.allclose(qtc[2, 2:8].table[0], range(2, 8)*u.m)

    assert qtc.names == ['x', 'y']
    assert qtc.physical_types == ['pos:x', 'pos:y']

    assert qtc.frame.axes_names == ('x', 'y')
    assert qtc.frame.axis_physical_types == ('custom:pos:x', 'custom:pos:y')


@pytest.mark.xfail(reason=">1D Tables not supported")
def test_slicing_quantity_table_coordinate_2d():
    qtc = QuantityTableCoordinate(*np.mgrid[0:10, 0:10]*u.m, mesh=False,
                                  names=['x', 'y'], physical_types=['pos:x', 'pos:y'])

    assert u.allclose(qtc[2:8, 2:8].table[0], (np.mgrid[2:8, 2:8]*u.m)[0])
    assert u.allclose(qtc[2:8, 2:8].table[1], (np.mgrid[2:8, 2:8]*u.m)[1])
    assert qtc.names == ['x', 'y']
    assert qtc.physical_types == ['pos:x', 'pos:y']

    assert qtc.frame.axes_names == ('x', 'y')
    assert qtc.frame.axis_physical_types == ('custom:pos:x', 'custom:pos:y')

    assert u.allclose(qtc[2, 2:8].table[0], 2*u.m)
    assert u.allclose(qtc[2, 2:8].table[1], (np.mgrid[2:8, 2:8]*u.m)[1])


def _assert_skycoord_equal(sc1, sc2):
    sc2 = sc2.transform_to(sc1.frame)

    assert sc1.shape == sc2.shape

    components1 = tuple(getattr(sc1.data, comp) for comp in sc1.data.components)
    components2 = tuple(getattr(sc2.data, comp) for comp in sc2.data.components)

    for c1, c2 in zip(components1, components2):
        assert u.allclose(c1, c2)


def test_slicing_skycoord_table_coordinate():
    # 1D, no mesh
    sc = SkyCoord(range(10)*u.deg, range(10)*u.deg)
    stc = SkyCoordTableCoordinate(sc, mesh=False, names=['lon', 'lat'], physical_types=['pos:x', 'pos:y'])

    _assert_skycoord_equal(stc[2:8].table, sc[2:8])
    _assert_skycoord_equal(stc[2].table, sc[2])
    assert stc.names == ['lon', 'lat']
    assert stc.physical_types == ['pos:x', 'pos:y']

    assert stc.frame.axes_names == ('lon', 'lat')
    assert stc.frame.axis_physical_types == ('custom:pos:x', 'custom:pos:y')

    # 2D, no mesh
    sc = SkyCoord(*np.mgrid[0:10, 0:10]*u.deg)
    stc = SkyCoordTableCoordinate(sc, mesh=False)
    _assert_skycoord_equal(stc[2:8, 2:8].table, sc[2:8, 2:8])
    _assert_skycoord_equal(stc[2, 2:8].table, sc[2, 2:8])


def test_slicing_time_table_coordinate():
    data = Time(["2011-01-01T00:00:00",
                 "2011-01-01T00:00:10",
                 "2011-01-01T00:00:20",
                 "2011-01-01T00:00:30"], format="isot")

    ttc = TimeTableCoordinate(data)
    assert (ttc.table == data).all()

    assert (ttc[2:8].table == data[2:8]).all()
    assert ttc[2].table == data[2]


def test_1d_distance_slice(lut_1d_distance):
    sub_ltc = lut_1d_distance[0:5]
    assert len(sub_ltc.table[0]) == 5


def test_3d_distance_slice(lut_3d_distance_mesh):
    sub_ltc = lut_3d_distance_mesh[0:5, 0:6, 0:7]
    assert len(sub_ltc.table[0]) == 5
    assert len(sub_ltc.table[1]) == 6
    assert len(sub_ltc.table[2]) == 7


@pytest.mark.xfail(reason=">1D Tables not supported")
def test_2d_nout_1_no_mesh_slice(lut_2d_distance_no_mesh):
    ltc = lut_2d_distance_no_mesh
    sub_ltc = ltc[0:2, 0:2]
    assert sub_ltc.table[0].shape == (2, 2)
    assert sub_ltc.table[1].shape == (2, 2)


def test_1d_skycoord_no_mesh_slice(lut_1d_skycoord_no_mesh):
    sub_ltc = lut_1d_skycoord_no_mesh[0:4]
    assert sub_ltc.table.shape == (4, )
    assert sub_ltc.table.shape == (4, )


@pytest.mark.xfail(reason=">1D Tables not supported")
def test_2d_skycoord_mesh_slice(lut_2d_skycoord_mesh):
    sub_ltc = lut_2d_skycoord_mesh[4:10, 5:10]
    assert sub_ltc.table.shape == (10,)
    assert sub_ltc._slice == [slice(4, 10, None), slice(5, 10, None)]

    assert sub_ltc.wcs.world_to_pixel(4*u.deg, 5*u.deg) == [0.0, 0.0]
    assert sub_ltc[1:, 1:].wcs.world_to_pixel(5*u.deg, 6*u.deg) == [0.0, 0.0]


@pytest.mark.xfail(reason=">1D Tables not supported")
def test_2d_skycoord_no_mesh_slice(lut_2d_skycoord_no_mesh):
    sub_ltc = lut_2d_skycoord_no_mesh[1:3, 1:2]
    assert sub_ltc.table.shape == (2, 1)


def test_1d_time_slice(lut_1d_time):
    sub_ltc = lut_1d_time[1:3]
    assert sub_ltc.table.shape == (2,)


def test_join_slice(lut_1d_time, lut_1d_wave):
    ltc = lut_1d_time & lut_1d_wave

    sub_ltc = ltc[2:8, 2:8]
    assert len(sub_ltc._table_coords) == 2
    assert (sub_ltc._table_coords[0].table == lut_1d_time.table[2:8]).all()
    assert u.allclose(sub_ltc._table_coords[1].table[0], lut_1d_wave.table[0][2:8])


def test_slicing_errors(lut_1d_time, lut_1d_wave, lut_1d_distance, lut_2d_skycoord_mesh):
    with pytest.raises(ValueError) as ei:
        lut_1d_time[1, 2]
    assert "slice with incorrect length" in str(ei)

    with pytest.raises(ValueError) as ei:
        lut_1d_wave[1, 2]
    assert "slice with incorrect length" in str(ei)

    with pytest.raises(ValueError) as ei:
        lut_1d_distance[1, 2]
    assert "slice with incorrect length" in str(ei)

    with pytest.raises(ValueError) as ei:
        lut_2d_skycoord_mesh[1, 2, 3]
    assert "slice with incorrect length" in str(ei)

    join = lut_1d_time & lut_1d_distance

    with pytest.raises(ValueError) as ei:
        join[1]
    assert "length of the slice" in str(ei)


def test_mtc_dropped_table(lut_1d_time):
    mtc = MultipleTableCoordinate(lut_1d_time)
    sub = mtc[0]

    assert len(sub._table_coords) == 0
    assert len(sub._dropped_coords) == 1

    dwd = sub.dropped_world_dimensions
    assert isinstance(dwd, dict)
    wao_classes = dwd.pop("world_axis_object_classes")
    assert all(isinstance(value, list) for value in dwd.values())
    assert all(len(value) == 1 for value in dwd.values())

    assert dwd["world_axis_names"] == ["time"]
    assert dwd["world_axis_units"] == ["s"]
    assert dwd["world_axis_physical_types"] == ["time"]
    assert dwd["world_axis_object_components"][0][0:2] == ("temporal", 0)
    assert wao_classes["temporal"][0] is Time
    assert dwd["value"] == [0*u.s]


def test_mtc_dropped_table_join(lut_1d_time, lut_2d_skycoord_mesh):
    mtc = MultipleTableCoordinate(lut_1d_time, lut_2d_skycoord_mesh)
    sub = mtc[0, :, :]

    assert len(sub._table_coords) == 1
    assert len(sub._dropped_coords) == 1

    dwd = sub.dropped_world_dimensions
    assert isinstance(dwd, dict)
    wao_classes = dwd.pop("world_axis_object_classes")
    assert all(isinstance(value, list) for value in dwd.values())
    assert all(len(value) == 1 for value in dwd.values())

    assert dwd["world_axis_names"] == ["time"]
    assert all(isinstance(u, str) for u in dwd["world_axis_units"])
    assert dwd["world_axis_units"] == ["s"]
    assert dwd["world_axis_physical_types"] == ["time"]
    assert dwd["world_axis_object_components"][0][0:2] == ("temporal", 0)
    assert wao_classes["temporal"][0] is Time
    assert dwd["value"] == [0*u.s]


def test_mtc_dropped_table_skycoord_join(lut_1d_time, lut_2d_skycoord_mesh):
    mtc = MultipleTableCoordinate(lut_1d_time, lut_2d_skycoord_mesh)
    sub = mtc[:, 0, 0]

    assert len(sub._table_coords) == 1
    assert len(sub._dropped_coords) == 1

    dwd = sub.dropped_world_dimensions
    assert isinstance(dwd, dict)
    wao_classes = dwd.pop("world_axis_object_classes")
    assert all(isinstance(value, list) for value in dwd.values())
    assert all(len(value) == 2 for value in dwd.values())

    assert dwd["world_axis_names"] == ["lon", "lat"]
    assert all(isinstance(u, str) for u in dwd["world_axis_units"])
    assert dwd["world_axis_units"] == ["deg", "deg"]
    assert dwd["world_axis_physical_types"] == ["pos.eq.ra", "pos.eq.dec"]
    assert dwd["world_axis_object_components"] == [("celestial", 0, "spherical.lon"), ("celestial", 1, "spherical.lat")]
    assert wao_classes["celestial"][0] is SkyCoord
    assert dwd["value"] == [0*u.deg, 0*u.deg]


@pytest.mark.xfail(reason=">1D Tables not supported")
def test_mtc_dropped_quantity_table(lut_1d_time, lut_2d_distance_no_mesh):
    mtc = MultipleTableCoordinate(lut_1d_time, lut_2d_distance_no_mesh)
    sub = mtc[:, 0, 0]

    assert len(sub._table_coords) == 1
    assert len(sub._dropped_coords) == 1

    pytest.importorskip("gwcs", minversion="0.16.2a1.dev17")

    dwd = sub.dropped_world_dimensions
    assert isinstance(dwd, dict)
    wao_classes = dwd.pop("world_axis_object_classes")
    assert all(isinstance(value, list) for value in dwd.values())
    assert dwd
    assert all(len(value) == 2 for value in dwd.values())

    assert dwd["world_axis_names"] == ["", ""]
    assert all(isinstance(u, str) for u in dwd["world_axis_units"])
    assert dwd["world_axis_units"] == ["km", "km"]
    assert dwd["world_axis_physical_types"] == ["custom:SPATIAL", "custom:SPATIAL"]
    assert dwd["world_axis_object_components"] == [("SPATIAL", 0, "value"), ("SPATIAL1", 0, "value")]
    assert wao_classes["SPATIAL"][0] is u.Quantity
    assert wao_classes["SPATIAL1"][0] is u.Quantity
    assert dwd["value"] == [0*u.km, 9*u.km]


def test_mtc_dropped_quantity_inside_table(lut_3d_distance_mesh):
    sub = lut_3d_distance_mesh[:, 0, :]

    assert len(sub.table) == 2

    pytest.importorskip("gwcs", minversion="0.16.2a1.dev17")

    dwd = sub.dropped_world_dimensions
    assert isinstance(dwd, dict)
    dwd.pop("world_axis_object_classes")
    assert all(isinstance(value, list) for value in dwd.values())
    assert dwd
    assert all(len(value) == 1 for value in dwd.values())

    sub = lut_3d_distance_mesh[:, 0, 0]

    assert len(sub.table) == 1

    dwd = sub.dropped_world_dimensions
    assert isinstance(dwd, dict)
    dwd.pop("world_axis_object_classes")
    assert all(isinstance(value, list) for value in dwd.values())
    assert dwd
    assert all(len(value) == 2 for value in dwd.values())


@pytest.mark.xfail(reason=">1D Tables not supported")
def test_mtc_dropped_quantity_inside_table_no_mesh(lut_2d_distance_no_mesh):
    """
    When not meshing, we don't drop a coord, as the coordinate for the sliced
    out axis can still vary along the remaining coordinate.
    """
    sub = lut_2d_distance_no_mesh[:, 0]

    assert len(sub.table) == 2

    pytest.importorskip("gwcs", minversion="0.16.2a1.dev17")

    dwd = sub.dropped_world_dimensions
    assert isinstance(dwd, dict)
    dwd.pop("world_axis_object_classes")
    assert not dwd


def test_mtc_dropped_quantity_join_drop_table(lut_1d_time, lut_3d_distance_mesh):
    mtc = MultipleTableCoordinate(lut_1d_time, lut_3d_distance_mesh)
    sub = mtc[:, 0, :, :]

    assert len(sub._table_coords) == 2
    assert len(sub._dropped_coords) == 0

    pytest.importorskip("gwcs", minversion="0.16.2a1.dev17")

    dwd = sub.dropped_world_dimensions
    assert isinstance(dwd, dict)
    dwd.pop("world_axis_object_classes")
    assert all(isinstance(value, list) for value in dwd.values())
    assert all(len(value) == 1 for value in dwd.values())

    sub = mtc[0, 0, :, :]

    assert len(sub._table_coords) == 1
    assert len(sub._dropped_coords) == 1

    pytest.importorskip("gwcs", minversion="0.16.2a1.dev17")

    dwd = sub.dropped_world_dimensions
    assert isinstance(dwd, dict)
    dwd.pop("world_axis_object_classes")
    assert all(isinstance(value, list) for value in dwd.values())
    assert all(len(value) == 2 for value in dwd.values())


################################################################################
# Tests of & operator
################################################################################


def test_and_base_table_coordinate():
    data = Time(["2011-01-01T00:00:00",
                 "2011-01-01T00:00:10",
                 "2011-01-01T00:00:20",
                 "2011-01-01T00:00:30"], format="isot")

    ttc = TimeTableCoordinate(data)

    qtc = QuantityTableCoordinate(range(10)*u.m, mesh=False)

    join = ttc & ttc
    assert isinstance(join, MultipleTableCoordinate)

    join2 = join & qtc

    assert isinstance(join2, MultipleTableCoordinate)
    assert len(join2._table_coords) == 3
    assert join2._table_coords[2] is qtc

    join3 = qtc & join

    assert isinstance(join3, MultipleTableCoordinate)
    assert len(join3._table_coords) == 3
    assert join3._table_coords[0] is qtc

    join4 = ttc & qtc
    assert isinstance(join4, MultipleTableCoordinate)
    assert len(join4._table_coords) == 2
    assert join4._table_coords[0] is ttc
    assert join4._table_coords[1] is qtc

    join5 = join & join
    assert isinstance(join5, MultipleTableCoordinate)
    assert len(join5._table_coords) == 4


def test_and_errors():
    data = Time(["2011-01-01T00:00:00",
                 "2011-01-01T00:00:10",
                 "2011-01-01T00:00:20",
                 "2011-01-01T00:00:30"], format="isot")

    ttc = TimeTableCoordinate(data)

    qtc = QuantityTableCoordinate(range(10)*u.m, mesh=False)

    with pytest.raises(TypeError) as ei:
        ttc & 5
    assert "unsupported operand type(s) for &: 'TimeTableCoordinate' and 'int'" in str(ei)

    join = ttc & qtc

    with pytest.raises(TypeError) as ei:
        join & 5
    assert "unsupported operand type(s) for &: 'MultipleTableCoordinate' and 'int'" in str(ei)

    with pytest.raises(TypeError) as ei:
        5 & join
    assert "unsupported operand type(s) for &: 'int' and 'MultipleTableCoordinate'" in str(ei)




from astropy.nddata.mixins.ndslicing import NDSlicingMixin
from astropy.wcs.wcsapi.wrappers.sliced_wcs import sanitize_slices

__all__ = ['NDCubeSlicingMixin']


class NDCubeSlicingMixin(NDSlicingMixin):
    # Inherit docstring from parent class
    __doc__ = NDSlicingMixin.__doc__

    def __getitem__(self, item):
        """
        Override the parent class method to explicitly catch `None` indices.

        This method calls ``_slice`` and then constructs a new object
        using the kwargs returned by ``_slice``.
        """
        if item is None or (isinstance(item, tuple) and None in item):
            raise IndexError("None indices not supported")

        item = tuple(sanitize_slices(item, len(self.dimensions)))
        sliced_cube = super().__getitem__(item)

        sliced_cube._global_coords._internal_coords = self.global_coords._internal_coords
        sliced_cube._extra_coords = self.extra_coords[item]

        return sliced_cube



import re

import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.tests.helper import assert_quantity_allclose
from astropy.units import Quantity
from astropy.wcs.wcsapi import HighLevelWCSWrapper
from numpy.testing import assert_allclose, assert_equal

from ndcube.wcs.wrappers import ReorderedLowLevelWCS


@pytest.fixture
def spectral_wcs(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def celestial_wcs(request):
    return request.getfixturevalue(request.param)


EXPECTED_SPECTRAL_CUBE_REPR = """
ReorderedLowLevelWCS Transformation

This transformation has 3 pixel and 3 world dimensions

Array shape (Numpy order): (6, 3, 7)

Pixel Dim  Axis Name  Data size  Bounds
        0  None               7  (1, 7)
        1  None               3  (1, 2.5)
        2  None               6  (-1, 5)

World Dim  Axis Name        Physical Type  Units
        0  Frequency        em.freq        Hz
        1  Right Ascension  pos.eq.ra      deg
        2  Declination      pos.eq.dec     deg

Correlation between pixel and world axes:

             Pixel Dim
World Dim    0    1    2
        0   no  yes   no
        1  yes   no  yes
        2  yes   no  yes
""".strip()


def test_spectral_cube(spectral_cube_3d_fitswcs):

    wcs = ReorderedLowLevelWCS(spectral_cube_3d_fitswcs,
                               pixel_order=[1, 2, 0],
                               world_order=[2, 0, 1])

    assert wcs.pixel_n_dim == 3
    assert wcs.world_n_dim == 3
    assert tuple(wcs.world_axis_physical_types) == ('em.freq', 'pos.eq.ra', 'pos.eq.dec')
    assert tuple(wcs.world_axis_units) == ('Hz', 'deg', 'deg')
    assert tuple(wcs.pixel_axis_names) == ('', '', '')
    assert tuple(wcs.world_axis_names) == ('Frequency',
                                           'Right Ascension',
                                           'Declination')
    assert_equal(wcs.axis_correlation_matrix, np.array([[0, 1, 0],
                                                        [1, 0, 1],
                                                        [1, 0, 1]]))

    assert wcs.pixel_shape == (7, 3, 6)
    assert wcs.array_shape == (6, 3, 7)
    assert wcs.pixel_bounds == ((1, 7), (1, 2.5), (-1, 5))

    pixel_scalar = (1.3, 2.3, 4.3)
    world_scalar = (-1.91e10, 5.4, -9.4)
    assert_allclose(wcs.pixel_to_world_values(*pixel_scalar), world_scalar)
    assert_allclose(wcs.array_index_to_world_values(*pixel_scalar[::-1]), world_scalar)
    assert_allclose(wcs.world_to_pixel_values(*world_scalar), pixel_scalar)
    assert_allclose(wcs.world_to_array_index_values(*world_scalar), [4, 2, 1])

    pixel_array = (np.array([1.3, 1.4]),
                   np.array([2.3, 2.4]),
                   np.array([4.3, 4.4]))
    world_array = (np.array([-1.91e10, -1.88e10]),
                   np.array([5.4, 5.2]),
                   np.array([-9.4, -9.2]))
    assert_allclose(wcs.pixel_to_world_values(*pixel_array), world_array)
    assert_allclose(wcs.array_index_to_world_values(*pixel_array[::-1]), world_array)
    assert_allclose(wcs.world_to_pixel_values(*world_array), pixel_array)
    assert_allclose(wcs.world_to_array_index_values(*world_array),
                    [[4, 4], [2, 2], [1, 1]])

    wcs_hl = HighLevelWCSWrapper(wcs)

    spectral, celestial = wcs_hl.pixel_to_world(*pixel_scalar)
    assert isinstance(spectral, Quantity)
    assert_quantity_allclose(spectral, world_scalar[0] * u.Hz)
    assert isinstance(celestial, SkyCoord)
    assert_quantity_allclose(celestial.ra, world_scalar[1] * u.deg)
    assert_quantity_allclose(celestial.dec, world_scalar[2] * u.deg)

    spectral, celestial = wcs_hl.pixel_to_world(*pixel_array)
    assert isinstance(spectral, Quantity)
    assert_quantity_allclose(spectral, world_array[0] * u.Hz)
    assert isinstance(celestial, SkyCoord)
    assert_quantity_allclose(celestial.ra, world_array[1] * u.deg)
    assert_quantity_allclose(celestial.dec, world_array[2] * u.deg)

    assert str(wcs) == EXPECTED_SPECTRAL_CUBE_REPR
    assert EXPECTED_SPECTRAL_CUBE_REPR in repr(wcs)


@pytest.mark.parametrize('order', [(1,), (1, 2, 2), (0, 1, 2, 3)])
def test_invalid(spectral_cube_3d_fitswcs, order):

    with pytest.raises(ValueError, match=re.escape('pixel_order should be a permutation of [0, 1, 2]')):
        ReorderedLowLevelWCS(spectral_cube_3d_fitswcs,
                             pixel_order=order,
                             world_order=[2, 0, 1])

    with pytest.raises(ValueError, match=re.escape('world_order should be a permutation of [0, 1, 2]')):
        ReorderedLowLevelWCS(spectral_cube_3d_fitswcs,
                             pixel_order=[1, 2, 0],
                             world_order=order)



import astropy.units as u

__all__ = ['unique_sorted']


def unique_sorted(iterable):
    """
    Return unique values in the order they are first encountered in the iterable.
    """
    lookup = set()  # a temporary lookup set
    return [ele for ele in iterable if ele not in lookup and lookup.add(ele) is None]


def convert_quantities_to_units(coords, units):
    """Converts a sequence of Quantities to units used in the WCS.

    Non-Quantity types in the sequence are allowed and ignored.

    Parameters
    ----------
    coords: iterable of `astropy.units.Quantity` or `None`
        The coordinates to be converted.

    units: iterable of `astropy.units.Unit` or `str`
        The units to which the coordinates should be converted.

    Returns
    -------
    converted_coords: iterable of `astropy.units.Quantity` or `None`
        The coordinates converted to the units.
        Non-quantity types remain.
    """
    return [coord.to(unit) if isinstance(coord, u.Quantity) else coord
            for coord, unit in zip(coords, units)]



"""
This module provides helpers introduced in https://github.com/astropy/astropy/pull/11950.
"""
from collections import OrderedDict, defaultdict

from astropy.wcs.wcsapi.high_level_api import default_order, rec_getattr

__all__ = ['values_to_high_level_objects', 'high_level_objects_to_values']

try:
    from astropy.wcs.wcsapi.high_level_api import high_level_objects_to_values, values_to_high_level_objects

except ImportError:

    def high_level_objects_to_values(*world_objects, low_level_wcs):
        """
        Convert the input high level object to low level values.

        This function uses the information in ``wcs.world_axis_object_classes`` and
        ``wcs.world_axis_object_components`` to convert the high level objects
        (such as `~.SkyCoord`) to low level "values" `~.Quantity` objects.

        This is used in `.HighLevelWCSMixin.world_to_pixel`, but provided as a
        separate function for use in other places where needed.

        Parameters
        ----------
        *world_objects: object
            High level coordinate objects.

        low_level_wcs: `.BaseLowLevelWCS`
            The WCS object to use to interpret the coordinates.
        """
        # Cache the classes and components since this may be expensive
        serialized_classes = low_level_wcs.world_axis_object_classes
        components = low_level_wcs.world_axis_object_components

        # Deserialize world_axis_object_classes using the default order
        classes = OrderedDict()
        for key in default_order(components):
            if low_level_wcs.serialized_classes:
                classes[key] = deserialize_class(serialized_classes[key],
                                                 construct=False)
            else:
                classes[key] = serialized_classes[key]

        # Check that the number of classes matches the number of inputs
        if len(world_objects) != len(classes):
            raise ValueError("Number of world inputs ({}) does not match "
                             "expected ({})".format(len(world_objects), len(classes)))

        # Determine whether the classes are uniquely matched, that is we check
        # whether there is only one of each class.
        world_by_key = {}
        unique_match = True
        for w in world_objects:
            matches = []
            for key, (klass, *_) in classes.items():
                if isinstance(w, klass):
                    matches.append(key)
            if len(matches) == 1:
                world_by_key[matches[0]] = w
            else:
                unique_match = False
                break

        # If the match is not unique, the order of the classes needs to match,
        # whereas if all classes are unique, we can still intelligently match
        # them even if the order is wrong.

        objects = {}

        if unique_match:

            for key, (klass, args, kwargs, *rest) in classes.items():

                if len(rest) == 0:
                    klass_gen = klass
                elif len(rest) == 1:
                    klass_gen = rest[0]
                else:
                    raise ValueError("Tuples in world_axis_object_classes should have length 3 or 4")

                # FIXME: For now SkyCoord won't auto-convert upon initialization
                # https://github.com/astropy/astropy/issues/7689
                from astropy.coordinates import SkyCoord
                if isinstance(world_by_key[key], SkyCoord):
                    if 'frame' in kwargs:
                        objects[key] = world_by_key[key].transform_to(kwargs['frame'])
                    else:
                        objects[key] = world_by_key[key]
                else:
                    objects[key] = klass_gen(world_by_key[key], *args, **kwargs)

        else:

            for ikey, key in enumerate(classes):

                klass, args, kwargs, *rest = classes[key]

                if len(rest) == 0:
                    klass_gen = klass
                elif len(rest) == 1:
                    klass_gen = rest[0]
                else:
                    raise ValueError("Tuples in world_axis_object_classes should have length 3 or 4")

                w = world_objects[ikey]
                if not isinstance(w, klass):
                    raise ValueError("Expected the following order of world "
                                     "arguments: {}".format(', '.join([k.__name__ for (k, _, _) in classes.values()])))

                # FIXME: For now SkyCoord won't auto-convert upon initialization
                # https://github.com/astropy/astropy/issues/7689
                from astropy.coordinates import SkyCoord
                if isinstance(w, SkyCoord):
                    if 'frame' in kwargs:
                        objects[key] = w.transform_to(kwargs['frame'])
                    else:
                        objects[key] = w
                else:
                    objects[key] = klass_gen(w, *args, **kwargs)

        # We now extract the attributes needed for the world values
        world = []
        for key, _, attr in components:
            if callable(attr):
                world.append(attr(objects[key]))
            else:
                world.append(rec_getattr(objects[key], attr))

        return world

    def values_to_high_level_objects(*world_values, low_level_wcs):
        """
        Convert low level values into high level objects.

        This function uses the information in ``wcs.world_axis_object_classes`` and
        ``wcs.world_axis_object_components`` to convert the high level objects
        (such as `~.SkyCoord`) to low level "values" `~.Quantity` objects.

        This is used in `.HighLevelWCSMixin.pixel_to_world`, but provided as a
        separate function for use in other places where needed.

        Parameters
        ----------
        *world_values: object
            Low level, "values" representations of the world coordinates.

        low_level_wcs: `.BaseLowLevelWCS`
            The WCS object to use to interpret the coordinates.
        """
        # Cache the classes and components since this may be expensive
        components = low_level_wcs.world_axis_object_components
        classes = low_level_wcs.world_axis_object_classes

        # Deserialize classes
        if low_level_wcs.serialized_classes:
            classes_new = {}
            for key, value in classes.items():
                classes_new[key] = deserialize_class(value, construct=False)
            classes = classes_new

        args = defaultdict(list)
        kwargs = defaultdict(dict)

        for i, (key, attr, _) in enumerate(components):
            if isinstance(attr, str):
                kwargs[key][attr] = world_values[i]
            else:
                while attr > len(args[key]) - 1:
                    args[key].append(None)
                args[key][attr] = world_values[i]

        result = []

        for key in default_order(components):
            klass, ar, kw, *rest = classes[key]
            if len(rest) == 0:
                klass_gen = klass
            elif len(rest) == 1:
                klass_gen = rest[0]
            else:
                raise ValueError("Tuples in world_axis_object_classes should have length 3 or 4")
            result.append(klass_gen(*args[key], *ar, **kwargs[key], **kw))

        return result
